///////////////////////////////////////////////////////////////////////////////
// PARSE.CPP  by Gregory Kwok <gkwok@jps.net>
///////////////////////////////////////////////////////////////////////////////
// V1.0   06 Aug 1994: First version
// V1.3   24 Aug 1994: Added /M switch to specify minimum occurences
// V1.5   05 Oct 1994: Fixed hyphen treatment; chars and words counted better
// V1.5b  14 Mar 1995: Fixed original words; some words weren't being counted
// V2.0   18 Nov 1995: Uses binary trees instead of an array, MUCH faster
// V2.1   18 Nov 1995: Removed word length limit, faster character input
// V2.3   08 May 1996: Added some ordinary words, first word char is capitalized
// V2.4   10 Sep 1996: Fixed memory leak! Removed ordinary words from help screen
// V2.5   16 Sep 1996: Changed word count types to long; used binary read mode
// V2.5   26 Nov 1997: Recompiled with DJGPP 32-bit compiler
///////////////////////////////////////////////////////////////////////////////

// Following applies only to 16-bit compilation:
// Compile with the compact memory module for best performance /mc
// it lets a small amount of code (64K) address a lot of data (1MB)

// Include files
#include <conio.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// Typedefs and user-defined types
enum Boolean { False, True };                    // False = 0, True = 1
struct wnode {                                   // One node of the binary tree
  char*  word;                                   // The word itself
  wnode* left;
  wnode* right;
  int freq;                                      // How many times it occurs
} * root;                                        // We must have a root!

// Function prototypes
int readArgs(int argc, char* argv[]);            // Parses the command line
long countWords(void);                           // Gets the letters, words, sents. stats
void writeWords(void);                           // Copies from the file to memory
void sortWords(void);                            // Finds the highest frequency words and prints them
int findMax(int& max, wnode* where);             // Finds the maximum freq in the tree
int printAllMax(int max, wnode* where, int& n);  // Prints all nodes with freq==max
void resetAllMax(int max, wnode* where);         // Sets all nodes with freq==max to freq=-1
Boolean isbreaker(const unsigned char tc);       // Returns true if certain punctuation
Boolean isord(const char* str);                  // Returns true if str is "ordinary"

// Global variables (I know, I know...)
FILE* parseFile;                                 // Main input file
long numLetters;                                 // Number of printable chars
long wrdLetters;                                 // Number of letters that are in words
long numWords;                                   // Number of words
long numSents;                                   // Number of sentences
int mComm = 10;                                  // How many most common words to find
int minOcc = 2;                                  // How many times each must appear
Boolean addOrd = True;                           // Add "ordinary" words?
Boolean semBrk = False;                          // Semicolon breaks a sentence?

///////////////////////////////////////////////////////////////////////////////
// Here we go...
///////////////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
  readArgs(argc, argv);
  if (!(parseFile = fopen(argv[1], "rb"))) {
    printf("Error: Could not open file %s.\n", argv[1]);
    exit(2);                                     // 2 = File open error
  }

  countWords();                                  // Get figures for the next statement
  printf(
   "\nPrintable letters:   %ld"
   "\nWords:               %ld"
   "\nSentences:           %ld"
   "\nAvg. letters/word:   %.1f"
   "\nAvg. words/sentence: %.1f",
   numLetters,
   numWords,
   numSents,
   (numWords ? ((float)wrdLetters / (float)numWords) : 0.0),
   (numSents ? ((float)numWords   / (float)numSents) : 0.0)
  );

  // Print the most common words if user wanted to (default)
  if (mComm) {                                   // Passing /C0 saves beaucoup memory and time
    writeWords();                                // Copy each word into the big array
    sortWords();                                 // Print the most common words
  }

  printf("\n");
  fclose(parseFile);                             // Close the input file
  return 0;                                      // 0 = Normal shutdown
}

///////////////////////////////////////////////////////////////////////////////
// Parses (?!) the command line and sets the switches
///////////////////////////////////////////////////////////////////////////////
int readArgs(int argc, char* argv[])
{
  char arg[6];
  int nmc;                                       // N Most Common
  int nmo;                                       // N Minimum Occurances

  if (argc == 1 || !strcmp(argv[1], "/?") || !strcmp(argv[1], "?")) {
    puts(
     "\nPARSE V2.5 \xfe Gregory Kwok <gkwok@jps.net>"
     "\nSyntax:  PARSE file [/Cn] [/Mn] [/O] [/S]"
     "\n   file   The file to read and parse, with an extension if necessary"
     "\n   /C     Prints the <n> most common words, default n = 10"
     "\n          \"/C\" is equivalent to \"/C0\""
     "\n   /M     Specifies that only words that appear <n> times or more are"
     "\n          to be added to the common word list, default n = 2"
     "\n   /O     Do not add ordinary words to the common word list, default"
     "\n          off (ordinary words are added)"
     "\n   /S     Semicolons and colons break sentences, default off"
     "\n"
     "\nNormally the following characters break a sentence:"
     "\n   (tab) <CR> <LF> <EOF> (space) ! \" ( ) , - . / ? [ ] { }"
     "\nThese extended ASCII characters also break sentences:"
     "\n   0x9b: \x9b  0x9c: \x9c  0xa8: \xa8  0xad: \xad  0xc4: \xc4  0xff: _"
    );
    exit(1);                                     // 1 = User needed help
  }

  for (int p = 2; p < argc; p++) {               // Go through each argument
    if (argv[p][0] == '/' || argv[p][0] == '-') {     // Q.Is it a switch?
      strcpy(arg, strupr(argv[p] + 1));
      if (argv[p][1] == 'C') {                   // "/Cn" = print n most common words
        nmc = atoi(arg + 1);                     // Extract number
        if (nmc < 0 || nmc > 200) {              // Q.Is it out of range 0-200?
          puts("/Cn must be in range of 0 to 200.");
          exit(3);                               // 3 = Argument error
        }
        mComm = nmc;                             // Set the verified value
      }
      else if (argv[p][1] == 'M') {              // "/Mn" = set minimum frequency
        nmo = atoi(arg + 1);                     // Extract number
        if (nmo < 1 || nmo > 10000) {            // Q.Is it out of range 1-10000?
          puts("/Mn must be in range of 1 to 10000.");
          exit(3);                               // 3 = Argument error
        }
        minOcc = nmo;                            // Set the verified value
      }

      // Other switches:
      // "/O" = Ignore ordinary words, "/O-" = read them
      // "/S" = ; and : break sentences, "/S-" = ; and : do not
      else if (!strcmp(arg, "O"))  addOrd = False;
      else if (!strcmp(arg, "O-")) addOrd = True;
      else if (!strcmp(arg, "S"))  semBrk = True;
      else if (!strcmp(arg, "S-")) semBrk = False;
    }
  }

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
// Counts the number of characters, words, and sentences
///////////////////////////////////////////////////////////////////////////////
long countWords(void)
{
  numWords = numSents = 0;                       // Counts number of words and sentences, resp.
  numLetters = wrdLetters = 0L;                  // Number of printable characters
  unsigned char c;                               // Current character

  while (isbreaker(c = fgetc(parseFile)))        // Read past leading spaces, if any
    if (c >= ' ')
      numLetters++;

  ungetc(c, parseFile);                          // c must be a non-breaker, so put it back

  while (!feof(parseFile)) {                     // Keep going to end of file
    if ((c = fgetc(parseFile)) >= ' ')           // Q.Is it printable?
      numLetters++;                              // Y.Increase character count
    if (c == '\x09')                             // Q.Is it a tab?
      numLetters += 7;                           // Y.Increase overall by 8
    if (!isbreaker(c))                           // Q.Is it is a non-space character?
      wrdLetters++;                              // Y.Increase word character count

    if (isbreaker(c) && c != '-') {              // Q.Is it a punctuation that's not a hyphen?
      numWords++;                                // Y.Increase word count
      if (c == '.' || c == '!' || c == '?')      // Q.Is it a sentence breaker?
        numSents++;                              // Y.Increase sentence count
      else if (semBrk && (c == ';' || c == ':')) // Other sentence breakers
        numSents++;

      do
        if ((c = fgetc(parseFile)) >= ' ')       // Keep getting characters until we hit a letter
          numLetters++;                          // Increase letter count
      while (isbreaker(c) && !feof(parseFile));  // Until a punctuation (hyphens
                                                 // ..included), or EOF
      wrdLetters++;                              // Last character must have been a non-breaker
    }
  }

  numLetters--;                                  // Minus 1 for the EOF
  wrdLetters--;                                  // See above
  rewind(parseFile);                             // Reset EOF flag and go to beginning
  return numWords;
}

///////////////////////////////////////////////////////////////////////////////
// Copy the words from the file into a binary tree in memory
///////////////////////////////////////////////////////////////////////////////
void writeWords(void)
{
  const WordLen = 32;                            // Maximum word length

  long currWord = 0;                             // What word are we on?
  unsigned char c;                               // Current character
  int cmp;                                       // Result of alphabetic comparison
  Boolean added;                                 // Used in binary tree adding

  char thisWord[WordLen + 1];                    // The word being added
  char wordLoc;                                  // Use a string index to read word into memory
  wnode* leaf, * where;                          // The leaf itself, and where to add it
  printf("\n\n");

  while (isbreaker(c = fgetc(parseFile)))        // Read past leading spaces, if any
    if (c >= ' ')                                // This loop is similar in structure to the one in countWords()
      numLetters++;                              // ..It is duplicated here in case /C0 was passed
  ungetc(c, parseFile);                          // c must be a non-breaker, so put it back

  wordLoc = 0;
  thisWord[0] = '\0';                            // Set thisWord to ""
  while (!feof(parseFile)) {                     // Keep going to end of file
    c = fgetc(parseFile);                        // Read a character
    if (isbreaker(c) && c != '-') {              // Q.Is it a punctuation that's not a hyphen?
      thisWord[wordLoc] = '\0';                  // Terminate thisWord
      thisWord[0] = toupper(thisWord[0]);        // Convert first character to uppercase
      strlwr(thisWord + 1);                      // And convert rest of word to lowercase

      if (!addOrd && isord(thisWord))            // If ordinary and addOrd==False, do nothing
        ;                                        // Else we need to add it to the tree
      else if (root == NULL) {                   // Tree doesn't exist, so create it
        root = new wnode;
        root->word = strdup(thisWord);
        root->left = root->right = NULL;
        root->freq = 1;

        if (!root || !root->word) {              // Very little memory (unlikely)
          fclose(parseFile);
          puts("\rNo memory to create binary tree, try /C0.");
          exit(4);
        }
      }
      else {                                     // Tree already exists, create a leaf
        leaf = new wnode;
        leaf->word = strdup(thisWord);
        leaf->left = leaf->right = NULL;
        leaf->freq = 1;

        if (!leaf || !leaf->word) {              // Out of memory
          fclose(parseFile);
          puts("\rOut of memory creating binary tree, try /C0.");
          exit(4);
        }

        where = root;                            // Start searching at the root
        added = False;                           // We haven't added the leaf yet

        do {
          cmp = stricmp(leaf->word, where->word);     // Compare the leaf with this location
          if (cmp < 0) {                         // Leaf < present
            if (where->left)                     // ..Q.Can we go left?
              where = where->left;               // ..Y.Move left
            else {                               // ..N.Add the leaf here
              where->left = leaf;
              added = True;
            }
          }
          else if (cmp == 0) {                   // Leaf == present
            where->freq++;                       // ..Just increase its frequency
            free(leaf->word);
            delete leaf;                         // ..And kill the leaf (it is not added)
            added = True;
          }
          else {                                 // Leaf > present
            if (where->right)                    // ..Q.Can we go right?
              where = where->right;              // ..Y.Move right
            else {                               // ..N.Add the leaf here
              where->right = leaf;
              added = True;
            }
          }
        } while (!added);                        // Keep looping until the leaf is added
      }

      wordLoc = 0;                               // Reset thisWord and its index
      thisWord[0] = thisWord[2] = '\0';

      if (!(currWord++ % 500))                   // End current word, print status every 500 words
        printf("\rWorking... %.0f%%",
         (float)currWord / (float)numWords * 100.0);

      do
        c = fgetc(parseFile);                    // Keep getting characters until we hit a letter
      while (isbreaker(c) && !feof(parseFile));  // Until a punctuation (hyphens
                                                 // ..included), or EOF
      if (!feof(parseFile))                      // Q.Is there still more to go?
        fseek(parseFile, -1, SEEK_CUR);          // Y.Move back 1 to read the last char
    }
    else {                                       // Copy the char to the word
      thisWord[wordLoc++] = c;

      // This section added in v2.4 to fix memory leak with words > 32 characters
      // It's slow, but I can think of no other way; you *need* a comparison after adding each character!
      if (wordLoc == WordLen) {                  // Q.Reached maximum word length?
        while (!isbreaker(c) || c == '-')        // Just read the rest of the word without storing it
          c = fgetc(parseFile);
        ungetc(c, parseFile);                    // Back up one character
      }
    }
  }
}

///////////////////////////////////////////////////////////////////////////////
// Finds the highest frequency words and prints them
///////////////////////////////////////////////////////////////////////////////
void sortWords(void)
{
  int max = 0, t = 1;                            // Yes, t is a temporary...
  if (mComm > 1)                                 // More than 1 common word wanted?
    printf("\r%d most common words:", mComm);
  else                                           // User only wanted most common word
    printf("\rMost common word:");

  for (int p = 1; p <= mComm; p++, max = 0, t = p) {
    printAllMax(findMax(max, root), root, t);    // Find the max in the tree and print it
    if (max >= minOcc)
      printf(" (%d)", max);                      // Print the number of times it occured
    resetAllMax(max, root);                      // Remove the frequency from the tree
  }
}

///////////////////////////////////////////////////////////////////////////////
// Finds every occurence of freq==max in the tree and sets the freq=-1
///////////////////////////////////////////////////////////////////////////////
void resetAllMax(int max, wnode* where)
{
  if (where->left)
    resetAllMax(max, where->left);

  if (where->freq == max)
    where->freq = -1;

  if (where->right)
    resetAllMax(max, where->right);
}

///////////////////////////////////////////////////////////////////////////////
// Finds every occurence of freq==max in the tree and prints <word>
///////////////////////////////////////////////////////////////////////////////
int printAllMax(int max, wnode* where, int& n)
{
  if (where->left)
    printAllMax(max, where->left, n);

  if (where->freq == max) {
    if (n > 0) {                                 // Q.This is the first time we've printed?
      if (max < minOcc) {                        // Q.And we should print?
        if (n == 1)                              // Q.NO common words at all
          printf("\n(none \xf2 %d)", minOcc);
        return max;
      }

      putchar('\n');
      if (mComm > 1)                             // Q.User wanted >1 common words
        printf("%d) ", n);                       // Y.Print its rank
      printf(where->word);                       // Always print the word
      n = -1;                                    // So "x)" is not displayed in the future
    }
    else {                                       // Multiple words have freq==max
      if (n > 16 && wherey() >= 24) {            // Q.Are we at the bottom of the screen?
        printf("\n-- More --");                  // Y.Pause and press a key
        getch();
        clrscr();
      }

      printf(", %s", where->word);               // Print a comma before the word
    }
  }

  if (where->right)
    printAllMax(max, where->right, n);

  return max;
}

///////////////////////////////////////////////////////////////////////////////
// Searches the tree and finds the highest freq, returns this value
///////////////////////////////////////////////////////////////////////////////
int findMax(int& max, wnode* where)
{
  if (where->left)
    findMax(max, where->left);

  if (where->freq > max)
    max = where->freq;

  if (where->right)
    findMax(max, where->right);

  return max;
}

///////////////////////////////////////////////////////////////////////////////
// This list of characters determines what starts and ends a word
///////////////////////////////////////////////////////////////////////////////
Boolean isbreaker(const unsigned char tc)
{
  if (tc == '\x09' || tc == '\x0d' || tc == '\x0a' || tc == '\x1a'
   || tc == ' ' || tc == '!' || tc == '\"' || tc == '(' || tc == ')'
   || tc == ',' || tc == '-' || tc == '.' || tc == '/' || tc == ':'
   || tc == ';' || tc == '?' || tc == '[' || tc == ']' || tc == '{'
   || tc == '}'
   || tc == (unsigned char)'\x9b' || tc == (unsigned char)'\x9c'
   || tc == (unsigned char)'\xa8' || tc == (unsigned char)'\xad'
   || tc == (unsigned char)'\xc4' || tc == (unsigned char)'\xff')

    return True;                                 // Character breaks a word

  return False;
}

///////////////////////////////////////////////////////////////////////////////
// If /O was passed, this list of words is considered "ordinary"
// Derived from Lotus(R)'s Ami Pro(TM) 3.0 for Windows(R) **LTSFRAM1.DIC**
///////////////////////////////////////////////////////////////////////////////
Boolean isord(const char* str)
{
  static char* ordlist[] = { "a", "all", "an", "and", "are", "as", "at",
    "be", "been", "but", "by", "for", "from", "had", "has", "have", "he",
    "her", "him", "his", "if", "in", "is", "it", "its", "more", "no", "not",
    "of", "on", "one", "or", "out", "she", "so", "that", "the", "their",
    "there", "they", "this", "to", "was", "we", "were", "what", "when",
    "which", "will", "with", "would",

    // Added with V2.3:
    "any", "anybody", "anyone", "anything", "anyway", "anywhere", "about",
    "am", "can", "cannot", "can't", "could", "couldn't", "could've", "do",
    "don't", "every", "everybody", "everyone", "everything", "everywhere",
    "he'll", "he's", "he'd", "hers", "i", "i'll", "i'm", "it'll", "it's",
    "i've", "just", "like", "me", "mine", "my", "our", "ours", "she'll",
    "she's", "she'd", "should", "shouldn't", "should've", "theirs", "them",
    "they'll", "they're", "they've", "us", "we'll", "we're", "we've",
    "where", "who", "why", "won't", "wouldn't", "would've", "you", "you'll",
    "you're", "your", "yours", "you've"
  };

  for (int co = 0; co < sizeof(ordlist) / sizeof(ordlist[0]); co++)
    if (!stricmp(str, ordlist[co]))                   // Compare each word
      return True;

  return False;
}
